from abc import ABC, abstractmethod

import torch


class ContextAwareModule(ABC, torch.nn.Module):
    @property
    @abstractmethod
    def context(self) -> torch.Tensor: ...

    @property
    @abstractmethod
    def equivariant_step(self) -> torch.Tensor: ...

    @property
    @abstractmethod
    def min_input_shape(self) -> torch.Tensor: ...

    @property
    @abstractmethod
    def min_output_shape(self) -> torch.Tensor: ...

    @property
    @abstractmethod
    def dims(self) -> int: ...



from typing import Sequence

import torch

from .tem import ContextAwareModule


class Downsample(ContextAwareModule):
    """
    The Downsample class wraps a MaxPool layer with a limited set of arguments.
    It is designed to be used with the UNet class and provides satisfies
    the `ContextAwareModule` interface.

    :param dims: the number of dimensions (1, 2, or 3)
    :param downsample_factor: the downsample factor for each dimension
    """

    _context: torch.Tensor
    _equivariant_step: torch.Tensor
    _dims: int

    def __init__(self, dims: int, downsample_factor: Sequence[int] | int):
        super().__init__()

        self._dims = dims
        self._context = torch.tensor((0,) * dims)

        pool = {
            1: torch.nn.MaxPool1d,
            2: torch.nn.MaxPool2d,
            3: torch.nn.MaxPool3d,
        }[dims]

        if isinstance(downsample_factor, int):
            _downsample_factor = (downsample_factor,) * dims
        else:
            _downsample_factor = tuple(downsample_factor)

        self.down = pool(
            _downsample_factor,
            stride=_downsample_factor,
            padding=(0,) * dims,
        )

        self._equivariant_step = self._context + torch.tensor(_downsample_factor)

    @property
    def context(self) -> torch.Tensor:
        """
        The context is always 0 for downsampling.
        """
        return self._context

    @property
    def equivariant_step(self) -> torch.Tensor:
        """
        The invariant step is the downsample factor for each dimension.
        """
        return self._equivariant_step

    @property
    def min_input_shape(self) -> torch.Tensor:
        """
        `min_input_shape` is equal to the downsample factor for each dimension.
        """
        return self.min_output_shape * self.equivariant_step

    @property
    def min_output_shape(self) -> torch.Tensor:
        """
        `min_output_shape` is just 1 for each dimension.
        """
        return torch.tensor((1,) * self.dims)

    @property
    def dims(self) -> int:
        """
        The number of dimensions (1, 2, or 3).
        """
        return self._dims

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Apply the downsample operation to the input tensor.

        :param x: the input tensor
        """
        for d in range(1, self.dims + 1):
            if x.size()[-d] % self.equivariant_step[-d] != 0:
                raise RuntimeError(
                    "Can not downsample shape %s with factor %s, mismatch "
                    "in spatial dimension %d"
                    % (x.size(), self.equivariant_step, self.dims - d)
                )

        return self.down(x)


class Upsample(ContextAwareModule):
    """
    The Upsample class wraps a Upsample layer with a limited set of arguments.
    It is designed to be used with the UNet class and satisfies
    the `ContextAwareModule` interface.

    :param dims: the number of dimensions
    :param scale_factor: the upsample factor
    :param mode: the upsample mode (nearest, bilinear, etc.)
    """

    _dims: int

    def __init__(
        self,
        dims: int,
        scale_factor: Sequence[int] | int,
        mode: str = "nearest",
    ):
        super().__init__()

        self._dims = dims
        self._equivariant_step = torch.tensor((1,) * self.dims) / torch.tensor(
            scale_factor
        )
        scale_factor = (
            tuple(scale_factor) if not isinstance(scale_factor, int) else scale_factor
        )
        self.up = torch.nn.Upsample(scale_factor=scale_factor, mode=mode)

    @property
    def context(self) -> torch.Tensor:
        """
        `context` is always 0 for upsampling.
        """
        return torch.tensor((0,) * self.dims)

    @property
    def equivariant_step(self) -> torch.Tensor:
        """
        The invariant step is the inverse of the upsample factor for each dimension.
        A upsample factor of 4 means each pixel shift in the input shifts the output by 4 pixels.
        """
        return self._equivariant_step

    @property
    def min_input_shape(self) -> torch.Tensor:
        """
        `min_input_shape` is equal to 1 for each dimension.
        """
        return torch.tensor((1,) * self.dims)

    @property
    def min_output_shape(self) -> torch.Tensor:
        """
        `min_output_shape` is equal to the upsample factor for each dimension.
        """
        return (self.min_input_shape / self.equivariant_step).int()

    @property
    def dims(self) -> int:
        """
        The number of dimensions.
        """
        return self._dims

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Apply the upsample operation to the input tensor.

        :param x: the input tensor
        """
        return self.up(x)



import torch

from .conv_pass import ConvPass
from .scale import Downsample, Upsample
from .tem import ContextAwareModule


class UModule(ContextAwareModule):
    """
    The UModule class is an abstraction of a single layer
    of a UNet. It involves an input conv pass, a downsample,
    a lower block, an upsample, and an output conv pass.
    It is designed to be used with the UNet class and satisfies
    the `ContextAwareModule` interface.

    :param in_conv_pass: the input conv pass
    :param downsample: the downsample layer
    :param lower_block: the lower block
    :param upsample: the upsample layer
    :param out_conv_pass: the output
    """

    equivariance_context: torch.Tensor
    _dims: int
    _equivariant_step: torch.Tensor

    def __init__(
        self,
        in_conv_pass: ConvPass,
        downsample: Downsample,
        lower_block: ContextAwareModule,
        upsample: Upsample,
        out_conv_pass: ConvPass,
        _equivariance_context: torch.Tensor | None = None,
    ):
        super().__init__()
        self._dims = in_conv_pass.dims
        self._equivariant_step = downsample.equivariant_step * lower_block.equivariant_step
        assert torch.equal(
            downsample.equivariant_step * upsample.equivariant_step,
            torch.tensor((1,) * self.dims),
        ), "Down and Up sampling must have the same scale factor"

        self.equivariance_context = (
            _equivariance_context.long()
            if _equivariance_context is not None
            else torch.tensor((0,) * self.dims)
        )
        assert (
            self.dims
            == downsample.dims
            == lower_block.dims
            == upsample.dims
            == out_conv_pass.dims
        ), "All modules must have the same number of dimensions"

        self.in_conv_pass = in_conv_pass
        self.downsample = downsample
        self.lower_block = lower_block
        self.upsample = upsample
        self.out_conv_pass = out_conv_pass

    @property
    def dims(self) -> int:
        """
        The number of dimensions (1, 2, or 3).
        """
        return self._dims

    @property
    def context(self) -> torch.Tensor:
        """
        `in_conv_pass.context` + `downsample.equivariant_step * lower_block.context`
        + `out_conv_pass.context` + (Optional `equivariance_context`)

        The equivariance context is only added during evaluation and is used to make
        sure the network is translation equivariant for easy blockwise processing
        without tiling artifacts.
        """
        base_context = (
            self.in_conv_pass.context
            + self.downsample.equivariant_step * self.lower_block.context
            + self.out_conv_pass.context
        )

        if self.training:
            return base_context
        else:
            return base_context + self.equivariance_context

    @property
    def equivariant_step(self) -> torch.Tensor:
        """
        The invariant step is the product of the downsample factors.
        """
        return self._equivariant_step

    @property
    def min_input_shape(self) -> torch.Tensor:
        """
        The minimum input shape that this module can accept.
        """
        # Some details about the lower block
        lower_block_input_shape = self.lower_block.min_input_shape
        lower_block_context = self.lower_block.context
        lower_block_output_shape = lower_block_input_shape - lower_block_context

        # Upsample the lower block output shape and subtract our out conv context
        min_lower_output = (
            lower_block_output_shape / self.upsample.equivariant_step
        ).long()
        min_out = min_lower_output - self.out_conv_pass.context

        # min_out could be negative. We want it to be at least [1,1]
        min_expansion = torch.tensor([1] * self.dims) - min_out
        min_expansion[min_expansion < 0] = 0

        # we must round this value up to the next multiple of the invariant step
        min_expansion = (
            torch.ceil(min_expansion / self.equivariant_step) * self.equivariant_step
        ).long()

        # whats the minimum input shape of the lower block scaled by the downsample step
        min_lower_input = lower_block_input_shape * self.downsample.equivariant_step

        # now we just add the min_expansion term, and the context from the input conv pass
        min_input_shape = min_lower_input + min_expansion + self.in_conv_pass.context
        return min_input_shape

    @property
    def min_output_shape(self) -> torch.Tensor:
        """
        The minimum output shape that this module can produce.
        `min_input_shape` - `context`
        """
        return self.min_input_shape - self.context

    def set_equivariance_context(self, equivariance_context: torch.Tensor):
        """
        Set the equivariance context to be used during evaluation.

        :param equivariance_context: the equivariance context per dimension
        """
        self.equivariance_context = equivariance_context

    @torch.jit.export
    def crop(self, x: torch.Tensor, shape: torch.Tensor) -> torch.Tensor:
        """
        Center-crop x to match spatial dimensions given by shape.

        :param x: the input tensor
        :param shape: the target shape
        """

        x_shape = x.size()[2:]
        offset = (torch.tensor(x_shape) - shape) // 2
        for i, (o, s) in enumerate(zip(offset, shape)):
            x = torch.slice_copy(x, i + 2, o.item(), o.item() + s)
        return x

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Apply the UModule to the input tensor.

        :param x: the input tensor
        """
        # simple processing
        f_in = self.in_conv_pass(x)
        g_in = self.downsample(f_in)
        g_out = self.lower_block(g_in)
        f_in_up = self.upsample(g_out)

        # crop f_in and f_in_up to ensure translation equivariance
        target_shape = torch.tensor(f_in_up.size()[-self.dims :])
        if not self.training:
            target_shape = target_shape - self.equivariance_context
        f_in = self.crop(f_in, target_shape)
        f_in_up = self.crop(f_in_up, target_shape)
        f_in_cat = torch.cat([f_in, f_in_up], dim=1)

        # final conv pass
        y = self.out_conv_pass(f_in_cat)
        return y



import pytest
import torch

from tems import (
    ConvPass,
    Downsample,
    UModule,
    UNet,
    Upsample,
)


def identity(module: torch.nn.Module, tmp_path) -> torch.nn.Module:
    return module


def save_load(module: torch.nn.Module, tmp_path) -> torch.nn.Module:
    torch.save(module, tmp_path / "module.pt")
    return torch.load(tmp_path / "module.pt", weights_only=False)


def jit_script(module: torch.nn.Module, tmp_path) -> torch.nn.Module:
    module = torch.jit.script(module)
    return module


def jit_save_load(module: torch.nn.Module, tmp_path) -> torch.nn.Module:
    module = torch.jit.script(module)
    torch.jit.save(module, tmp_path / "module.jit")
    return torch.jit.load(tmp_path / "module.jit")


@pytest.mark.parametrize(
    "in_channels, out_channels, dims, kernel_sizes, activation, padding",
    [
        (1, 1, 2, [3], torch.nn.ReLU, "valid"),
        (2, 2, 3, [(3, 3, 3)], torch.nn.ReLU, "same"),
        (4, 4, 3, ((5, 5, 5), (3, 3, 3)), torch.nn.LeakyReLU, "valid"),
        (8, 8, 2, [7, [3, 3]], torch.nn.Sigmoid, "same"),
    ],
)
@pytest.mark.parametrize("serde", [identity, save_load, jit_script, jit_save_load])
def test_convpass(
    in_channels, out_channels, dims, kernel_sizes, activation, padding, tmp_path, serde
):
    # generate a few versions of the model (serialized/deserialized, jit, etc)
    conv_pass = ConvPass(
        dims=dims,
        in_channels=in_channels,
        out_channels=out_channels,
        kernel_sizes=kernel_sizes,
        activation=activation,
        padding=padding,
    )
    conv_pass_loaded = serde(conv_pass, tmp_path)

    # generate input data
    in_data = torch.rand(1, in_channels, *(conv_pass.context + 5))

    # compute output data
    out_data = conv_pass(in_data)
    out_data_loaded = conv_pass_loaded(in_data)

    # check that all versions of the model give the same output
    assert torch.equal(out_data, out_data_loaded)

    assert torch.equal(
        torch.tensor(in_data.shape[2:]) - torch.tensor(out_data.shape[2:]),
        conv_pass.context,
    ), (in_data.shape, out_data.shape, conv_pass.context)


@pytest.mark.parametrize(
    "dims, downsample_factor, invalid_input",
    [
        (3, (2, 3, 4), True),
        (2, 2, True),
        (3, 3, True),
        (2, [3, 1], True),
        (3, [2, 3, 4], False),
        (2, 2, False),
        (3, 3, False),
        (2, (3, 1), False),
    ],
)
@pytest.mark.parametrize("serde", [identity, save_load, jit_script, jit_save_load])
def test_downsample(dims, downsample_factor, invalid_input, serde, tmp_path):
    # build a few versions of the model (serialized/deserialized, jit, etc)
    downsample = Downsample(dims=dims, downsample_factor=downsample_factor)
    downsample_loaded = serde(downsample, tmp_path)

    # get the invariant step
    equivariant_step = downsample.equivariant_step
    if invalid_input:
        # generate invalid input data
        in_data = torch.rand(1, 1, *(equivariant_step + 1))

        # check that the models raise errors
        with pytest.raises(RuntimeError):
            downsample(in_data)
        with pytest.raises((RuntimeError, torch.jit.Error)):
            downsample_loaded(in_data)
    else:
        # generate valid input data
        in_data = torch.rand(1, 1, *(equivariant_step * 4))

        # process data
        out_data = downsample(in_data)
        out_data_loaded = downsample_loaded(in_data)

        # check results equal
        assert torch.equal(out_data, out_data_loaded)


@pytest.mark.parametrize(
    "dims, scale_factor, mode",
    [
        (1, 2, "nearest"),
        (3, 3, "trilinear"),
        (2, (4, 2), "bilinear"),
        (3, [2, 4, 3], "nearest"),
    ],
)
@pytest.mark.parametrize("serde", [identity, save_load, jit_script, jit_save_load])
def test_upsample(dims, scale_factor, mode, serde, tmp_path):
    # build a few versions of the model (serialized/deserialized, jit, etc)
    upsample = Upsample(
        dims=dims,
        scale_factor=scale_factor,
        mode=mode,
    )
    upsample_loaded = serde(upsample, tmp_path)

    # create in data
    in_data = torch.rand(1, 1, *((3,) * dims))

    # process data
    out_data = upsample(in_data)
    out_data_loaded = upsample_loaded(in_data)

    # check results equal
    assert torch.equal(out_data, out_data_loaded)


@pytest.mark.parametrize(
    "in_shape, in_conv_pass, downsample, lower_block, upsample, out_conv_pass, equivariance_context",
    [
        (
            (14, 20),
            ConvPass(dims=2, kernel_sizes=[3]),
            Downsample(dims=2, downsample_factor=(2, 3)),
            ConvPass(dims=2, kernel_sizes=[3]),
            Upsample(dims=2, scale_factor=(2, 3)),
            ConvPass(dims=2, in_channels=2, kernel_sizes=[3]),
            (2, 4),
        ),
    ],
)
@pytest.mark.parametrize("serde", [identity, save_load, jit_script, jit_save_load])
def test_umodule(
    in_shape,
    in_conv_pass,
    downsample,
    lower_block,
    upsample,
    out_conv_pass,
    equivariance_context,
    serde,
    tmp_path,
):
    # a few versions of the model (serialized/deserialized, jit, etc)
    umodule = UModule(
        in_conv_pass=in_conv_pass,
        downsample=downsample,
        lower_block=lower_block,
        upsample=upsample,
        out_conv_pass=out_conv_pass,
        _equivariance_context=torch.tensor(equivariance_context),
    )
    umodule_loaded = serde(umodule, tmp_path)

    # in data
    min_input_shape = umodule.min_input_shape
    min_output_shape = min_input_shape - umodule.context
    in_data = torch.rand(1, 1, *(min_input_shape))

    # out data versions
    out_data = umodule(in_data)
    out_data_loaded = umodule_loaded(in_data)

    # check that the data is the shape we expect
    assert out_data.shape[2:] == tuple(min_output_shape)

    # check that all versions of the model give the same output
    assert torch.equal(out_data, out_data_loaded)

    # check eval mode
    umodule = umodule.eval()
    umodule_loaded = umodule_loaded.eval()

    # generate input data
    in_data = torch.rand(1, 1, *(in_shape))

    # generate eval mode data
    out_data = umodule(in_data)
    out_data_loaded = umodule_loaded(in_data)

    # check that all versions of the model give the same output
    assert torch.equal(out_data, out_data_loaded)

    # check eval mode generated data with expected shape
    expected_output_shape = (
        torch.tensor(in_shape)
        - in_conv_pass.context
        - lower_block.context * downsample.equivariant_step
        - out_conv_pass.context
        - umodule.equivariance_context
    )
    assert out_data.shape[2:] == tuple(expected_output_shape)


@pytest.mark.parametrize(
    "dims, bottleneck, levels",
    [
        (
            2,
            ConvPass(2),
            [
                (
                    ConvPass(2),
                    Downsample(2, 2),
                    Upsample(2, 2),
                    ConvPass(2, in_channels=2),
                ),
                (
                    ConvPass(2),
                    Downsample(2, 2),
                    Upsample(2, 2),
                    ConvPass(2, in_channels=2),
                ),
                (
                    ConvPass(2),
                    Downsample(2, 2),
                    Upsample(2, 2),
                    ConvPass(2, in_channels=2),
                ),
                (
                    ConvPass(2),
                    Downsample(2, 2),
                    Upsample(2, 2),
                    ConvPass(2, in_channels=2),
                ),
            ],
        ),
    ],
)
@pytest.mark.parametrize("serde", [identity, save_load, jit_script, jit_save_load])
def test_unet(dims, bottleneck, levels, serde, tmp_path):
    unet = UNet(dims=dims, bottleneck=bottleneck, levels=levels)
    unet_loaded = serde(unet, tmp_path)

    # generate input data
    in_data = torch.rand(1, 1, *(unet.min_input_shape))

    # process data
    out_data = unet(in_data)
    out_data_loaded = unet_loaded(in_data)

    # check that all versions of the model give the same output
    assert torch.equal(out_data, out_data_loaded)

    # check eval mode
    unet = unet.eval()
    unet_loaded = unet_loaded.eval()

    # generate input data
    in_data = torch.rand(1, 1, *(unet.min_input_shape))

    # generate eval mode data
    out_data = unet(in_data)
    out_data_loaded = unet_loaded(in_data)

    # check that all versions of the model give the same output
    assert torch.equal(out_data, out_data_loaded)



from .conv_pass import ConvPass as ConvPass
from .scale import Downsample as Downsample
from .scale import Upsample as Upsample
from .umodule import UModule as UModule
from .unet import UNet as UNet

__version__ = "1.0.2"
__version_info__ = tuple(int(i) for i in __version__.split("."))



from itertools import product

import numpy as np
import pytest
import torch
from funlib.geometry import Coordinate, Roi
from funlib.persistence.arrays import Array

from tems import (
    UNet,
)


def gcd(a: int, b: int):
    """
    Compute the greatest common divisor of two integers.
    """
    while b:
        a, b = b, a % b
    return a


def gcd_coord(self: Coordinate, other: Coordinate):
    """
    Compute the greatest common divisor of two coordinates.
    """
    return Coordinate(*[gcd(a, b) for a, b in zip(self, other)])


Coordinate.gcd = gcd_coord


@pytest.mark.parametrize("eval_mode", [True, False])
@pytest.mark.parametrize(
    "downsample_factors",
    [
        [[2, 1], [2, 1], [2, 1], [2, 1]],
        [[2, 1], [3, 1], [4, 1], [2, 1]],
        [[2, 1], [4, 1], [3, 1], [5, 1]],
    ],
)
def test_translation_equivariance(eval_mode, downsample_factors):
    in_channels, out_channels = 1, 1
    unet = UNet.funlib_api(
        dims=2,
        in_channels=in_channels,
        num_fmaps=out_channels,
        fmap_inc_factor=1,
        downsample_factors=downsample_factors,
        activation="Identity",
    )
    if eval_mode:
        unet = unet.eval()
    min_output_shape = Coordinate(unet.min_output_shape)
    context = Coordinate(unet.context) // 2
    downsampling = Coordinate(unet.equivariant_step)
    blocks = Coordinate(1, 1) + downsampling / downsampling.gcd(
        min_output_shape
    ) * Coordinate(1, 0)

    in_array = Array(
        np.arange(
            (
                in_channels
                * min_output_shape[0]
                * min_output_shape[1]
                * blocks[0]
                * blocks[1]
            )
        ).reshape(1, in_channels, *(min_output_shape * blocks)),
        voxel_size=(1,) * len(min_output_shape),
    )
    blockwise_out_array = Array(
        np.zeros((1, out_channels, *(min_output_shape * blocks))),
        voxel_size=(1,) * len(min_output_shape),
    )
    full_out_array = Array(
        np.zeros((1, out_channels, *(min_output_shape * blocks))),
        voxel_size=(1,) * len(min_output_shape),
    )

    # process blockwise
    roi = Roi((0, 0), min_output_shape)

    for i, j in product(range(blocks[0]), range(blocks[1])):
        shift = Coordinate(i, j) * min_output_shape
        block_roi = roi + shift
        in_data = in_array.to_ndarray(block_roi.grow(context, context))
        out_data = unet(torch.from_numpy(in_data).float()).detach().cpu().numpy()
        blockwise_out_array[block_roi] = out_data

    full_out_array[full_out_array.roi] = (
        unet(
            torch.from_numpy(
                in_array.to_ndarray(in_array.roi.grow(context, context))
            ).float()
        )
        .detach()
        .cpu()
        .numpy()
    )

    a, b = blockwise_out_array[:], full_out_array[:]
    if not eval_mode:
        with pytest.raises(AssertionError):
            assert np.allclose(a, b), "Blockwise and in-memory output data do not match"
    else:
        assert np.allclose(a, b), "Blockwise and in-memory output data do not match"



from typing import Any, Sequence

import torch
from torch.nn import Conv1d, Conv2d, Conv3d

from .tem import ContextAwareModule


class ConvPass(ContextAwareModule):
    """
    The ConvPass class wraps a series of convolutional layers with a limited set
    of arguments. It is designed to be used with the UNet class and provides satisfies
    the `ContextAwareModule` interface.

    :param dims: the number of dimensions (1, 2, or 3)
    :param in_channels: the number of input channels
    :param out_channels: the number of output channels
    :param kernel_sizes: a sequence of kernel sizes for each convolutional layer
    :param activation: the activation function to use after each convolutional layer
    :param padding: the padding mode to use for the convolutional layers
    """

    _context: torch.Tensor
    _equivariant_step: torch.Tensor
    _dims: int

    def __init__(
        self,
        dims: int,
        in_channels: int = 1,
        out_channels: int = 1,
        kernel_sizes: Sequence[Sequence[int] | int] = (3, 3),
        activation: type[torch.nn.Module] = torch.nn.ReLU,
        padding: str = "valid",
    ):
        super().__init__()

        self._dims = dims
        self._context = torch.tensor((0,) * dims)
        self._equivariant_step = torch.tensor((1,) * dims)

        layers: list[torch.nn.Module] = []

        conv: Any = {
            1: Conv1d,
            2: Conv2d,
            3: Conv3d,
        }[dims]

        for kernel_size in kernel_sizes:
            # if isinstance(kernel_size, int):
            #     _kernel_size = (kernel_size,) * dims
            # else:
            #     _kernel_size = tuple(kernel_size)
            conv_layer = conv(
                in_channels,
                out_channels,
                kernel_size,
                padding=padding,
            )
            # initialize the weights properly
            assert conv_layer.bias is not None
            torch.nn.init.zeros_(conv_layer.bias)
            if activation is torch.nn.ReLU:
                torch.nn.init.kaiming_uniform_(conv_layer.weight, nonlinearity="relu")
            elif activation is torch.nn.LeakyReLU:
                torch.nn.init.kaiming_uniform_(
                    conv_layer.weight, nonlinearity="leaky_relu"
                )
            elif activation is torch.nn.Sigmoid or activation is torch.nn.Tanh:
                torch.nn.init.xavier_uniform_(conv_layer.weight)
            elif activation is torch.nn.Identity:
                torch.nn.init.constant_(conv_layer.weight, 1.0)
            layers.append(conv_layer)

            if padding == "valid":
                self._context += torch.tensor(kernel_size) - 1

            in_channels = out_channels

            if activation is not None:
                layers.append(activation())

        self.conv_pass = torch.nn.Sequential(*layers)

    @property
    def context(self) -> torch.Tensor:
        """
        The context of the ConvPass module.
        Set to zero if the padding is "same".
        Set to the sum of the [kernel_size - 1 for kernel_size in kernel_sizes] if the padding is "valid".
        """
        return self._context

    @property
    def equivariant_step(self) -> torch.Tensor:
        """
        Always 1 for ConvPass since we don't yet support strided convolutions.
        """
        return self._equivariant_step

    @property
    def min_input_shape(self) -> torch.Tensor:
        """
        Simply 1 + `self.context`.
        """
        return self.min_output_shape + self._context

    @property
    def min_output_shape(self) -> torch.Tensor:
        """
        Always 1
        """
        return torch.tensor((1,) * self.dims)

    @property
    def dims(self) -> int:
        """
        The number of dimensions of the ConvPass module.
        """
        return self._dims

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Apply a series of convolutions to the input tensor.

        :param x: the input tensor
        """
        return self.conv_pass(x)



from typing import Sequence

import torch

from .conv_pass import ConvPass
from .scale import Downsample, Upsample
from .tem import ContextAwareModule
from .umodule import UModule


class UNet(ContextAwareModule):
    """
    A UNet implementation with helper functions to calculate the
    minimum input and output shapes of the network, along with the
    context and appropriate step size for translation equivariance.

    This UNet is also fully scriptable with `torch.jit.script` making
    it easy to share and deploy.

    :param dims: the number of dimensions
    :param bottleneck: the bottleneck module
    :param levels: a sequence of tuples containing the left conv pass,
        downsample, upsample, and right conv pass for each level
    """

    _dims: int

    def __init__(
        self,
        dims: int,
        bottleneck: ContextAwareModule,
        levels: Sequence[tuple[ConvPass, Downsample, Upsample, ConvPass]],
    ):
        super().__init__()

        self._dims = dims

        head_module: ContextAwareModule | None = None
        for left, down, up, right in reversed(levels):
            head_module = UModule(
                in_conv_pass=left,
                downsample=down,
                lower_block=head_module if head_module is not None else bottleneck,
                upsample=up,
                out_conv_pass=right,
            )
        assert head_module is not None, "0 level UNet not supported"

        self.head_module = head_module

        # handle cropping to ensure translation equivariance
        # get all downsampling factors
        downsampling = [torch.tensor((1,) * dims)] + [
            downsample.equivariant_step for _, downsample, _, _ in levels
        ]
        stacked_downsampling = torch.stack(downsampling)
        layer_downsampling = torch.cumprod(stacked_downsampling, dim=0)
        total_downsampling = torch.prod(stacked_downsampling, dim=0)

        # get all layer output_sizes
        output_shape = head_module.min_output_shape

        # invariant output_shape: round output_shape down to the next multiple of total_downsampling
        invariant_output_shape = (
            torch.floor(output_shape / total_downsampling) * total_downsampling
        )
        to_crop = output_shape - invariant_output_shape

        crop_amounts: list[torch.Tensor] = []
        for downsample_factor in reversed(layer_downsampling[:-1]):
            can_crop = to_crop % (downsample_factor * 2) == 0
            crop_amount = (to_crop / downsample_factor) * can_crop
            crop_amounts = [crop_amount] + crop_amounts
            to_crop = to_crop - crop_amount * downsample_factor
        stacked_crop_amounts: torch.Tensor = torch.stack(crop_amounts)

        for crop_amount in stacked_crop_amounts:
            assert isinstance(head_module, UModule)
            head_module.set_equivariance_context(crop_amount.long())
            head_module = head_module.lower_block

    @property
    def dims(self) -> int:
        """
        The number of dimensions (1, 2, or 3).
        """
        return self._dims

    @property
    def context(self) -> torch.Tensor:
        """
        The context of the UNet.
        """
        return self.head_module.context

    @property
    def equivariant_step(self) -> torch.Tensor:
        """
        The invariant step is the product of all downsampling factors.
        """
        return self.head_module.equivariant_step

    @property
    def min_input_shape(self) -> torch.Tensor:
        """
        The minimum input shape that this module can accept.
        """
        return self.head_module.min_input_shape

    @property
    def min_output_shape(self) -> torch.Tensor:
        """
        The minimum output shape that this module can produce.
        """
        return self.head_module.min_output_shape

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Apply the UNet to the input tensor.

        :param x: the input tensor
        """
        return self.head_module(x)

    @classmethod
    def funlib_api(
        cls,
        dims: int,
        in_channels: int,
        num_fmaps: int,
        fmap_inc_factor: int,
        downsample_factors: Sequence[Sequence[int] | int],
        kernel_size_down: Sequence[Sequence[Sequence[int] | int]] | None = None,
        kernel_size_up: Sequence[Sequence[Sequence[int] | int]] | None = None,
        activation: str = "ReLU",
        num_fmaps_out: int | None = None,
        num_heads: int = 1,
        constant_upsample: bool = False,
        padding: str = "valid",
    ):
        """
        A helper method to match the API of the funlib UNet as closely as possible.
        This method is helpful to creating a UNet with a more compact API whereas
        the default constructor makes you define every layer explicitly.

        :param dims: the number of dimensions
        :param in_channels: the number of input channels
        :param num_fmaps: the number of feature maps
        :param fmap_inc_factor: the factor by which to increase the number of feature maps
        :param downsample_factors: the downsample factors for each level
        :param kernel_size_down: the kernel size for the downsample convolutions
        :param kernel_size_up: the kernel size for the upsample convolutions
        :param activation: the activation function to use
        :param num_fmaps_out: the number of output feature maps
        :param num_heads: the number of heads to use
        :param constant_upsample: whether to use constant upsampling
        :param padding: the padding mode to use. Supported values are "valid" and "same".
        """
        _activation: type[torch.nn.Module] = getattr(torch.nn, activation)
        kernel_size_up = (
            kernel_size_up
            if kernel_size_up is not None
            else [(3, 3)] * len(downsample_factors)
        )
        kernel_size_down = (
            kernel_size_down
            if kernel_size_down is not None
            else [(3, 3)] * len(downsample_factors)
        )
        if (
            kernel_size_down is not None
            and len(kernel_size_down) == len(downsample_factors) + 1
        ):
            bottleneck_kernel = kernel_size_down[-1]
        else:
            bottleneck_kernel = [3, 3]

        layers = []
        for i, (kernel_down, scale_factor, kernel_up) in enumerate(
            zip(
                kernel_size_down,
                downsample_factors,
                kernel_size_up,
            )
        ):
            layers.append(
                (
                    ConvPass(
                        dims,
                        in_channels
                        if i == 0
                        else num_fmaps * fmap_inc_factor ** (i - 1),
                        num_fmaps * fmap_inc_factor**i,
                        kernel_sizes=kernel_down,
                        activation=_activation,
                        padding=padding,
                    ),
                    Downsample(dims, scale_factor),
                    Upsample(dims, scale_factor),
                    ConvPass(
                        dims,
                        num_fmaps * fmap_inc_factor**i
                        + num_fmaps * fmap_inc_factor ** (i + 1),
                        num_fmaps * fmap_inc_factor**i,
                        kernel_sizes=kernel_up,
                        activation=_activation,
                        padding=padding,
                    ),
                )
            )
        bottleneck = ConvPass(
            dims,
            num_fmaps * fmap_inc_factor ** (len(downsample_factors) - 1),
            num_fmaps * fmap_inc_factor ** len(downsample_factors),
            kernel_sizes=bottleneck_kernel,
            activation=_activation,
            padding=padding,
        )
        return UNet(dims, bottleneck, layers)
